/*
** Copyright (c) 2005-2008, 2010-2011 7B Software, Inc.
** All rights reserved.
**
** This file is part of XMLP.
**
** XMLP is free software: you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation, either version 3 of the License, or
** (at your option) any later version.
**
** XMLP is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public License
** along with XMLP.  If not, see <http://www.gnu.org/licenses/>.
**
** Author contact information:
**   mark.smith@7bsoftware.com
**   http://www.7bsoftware.com
*/

/*
** This file implements the TCL interface to the XmlParser module.
*/
#include <tcl.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>
#include <inttypes.h>
#include "xmlptcl.h"

static int AttrCmdProc(Tcl_Interp*, int, Tcl_Obj *const*, XmlCursor*);

/*
** The following code does a unix "glob" style comparison of two strings.
** The pattern can contain wildcards as follows:  "*" matches zero or
** more characters.  "?" matches any single character.  "[...]" matches
** any one of the enclosed characters.  "[^...]" matches any character
** except those enclosed.
**
** This code is lifted directly from the SQLite library (with minor
** modifications).  SQLite is in the public domain.
*/
static int globUnsignedCompare(
  const unsigned char *zString,    // The string we are attempt to match
  const unsigned char *zPattern    // The pattern string
){
  register int c;
  int invert;
  int seen;
  int c2;
  // const unsigned char *zString = (const unsigned char*)zStr;
  // const unsigned char *zPattern = (const unsigned char*)zPat;
# define globNextChar(X) (++(X))
# define globCharVal(X)  ((int)*(X))

  while( (c = *zPattern)!=0 ){
    switch( c ){
      case '*':
        while( (c=zPattern[1]) == '*' || c == '?' ){
          if( c=='?' ){
            if( *zString==0 ) return 0;
            globNextChar(zString);
          }
          zPattern++;
        }
        if( c==0 ) return 1;
        if( c=='[' ){
          while( *zString && globUnsignedCompare(&zPattern[1],zString)==0 ){
            globNextChar(zString);
          }
          return *zString!=0;
        }else{
          while( (c2 = *zString)!=0 ){
            while( c2 != 0 && c2 != c ){ c2 = *++zString; }
            if( c2==0 ) return 0;
            if( globUnsignedCompare(&zPattern[1],zString) ) return 1;
            globNextChar(zString);
          }
          return 0;
        }
      case '?': {
        if( *zString==0 ) return 0;
        globNextChar(zString);
        zPattern++;
        break;
      }
      case '[': {
        int prior_c = 0;
        seen = 0;
        invert = 0;
        c = globCharVal(zString);
        if( c==0 ) return 0;
        c2 = *++zPattern;
        if( c2=='^' ){ invert = 1; c2 = *++zPattern; }
        if( c2==']' ){
          if( c==']' ) seen = 1;
          c2 = *++zPattern;
        }
        while( (c2 = globCharVal(zPattern))!=0 && c2!=']' ){
          if( c2=='-' && zPattern[1]!=']' && zPattern[1]!=0 && prior_c>0 ){
            zPattern++;
            c2 = globCharVal(zPattern);
            if( c>=prior_c && c<=c2 ) seen = 1;
            prior_c = 0;
          }else if( c==c2 ){
            seen = 1;
            prior_c = c2;
          }else{
            prior_c = c2;
          }
          globNextChar(zPattern);
        }
        if( c2==0 || (seen ^ invert)==0 ) return 0;
        globNextChar(zString);
        zPattern++;
        break;
      }
      default: {
        if( c != *zString ) return 0;
        zPattern++;
        zString++;
        break;
      }
    }
  }
  return *zString==0;
}
#define globCompare(A,B) \
   globUnsignedCompare((const unsigned char*)A,(const unsigned char*)B)

/*
** The user data for a TCL trace callback is a pointer to an instance
** of the following structure.
*/
class TraceCallbackData {
public:
  Tcl_Interp *interp;  // The interpreter in which to run the callback
  char *zCallback;     // Text of the callback script
  TraceCallbackData(Tcl_Interp *pInterp,const char *zProc){
     interp = pInterp;
     zCallback = new char[strlen(zProc)+1];
     if( zCallback ) strcpy(zCallback, zProc);
  };
  ~TraceCallbackData(){
     if( zCallback ){ delete [] zCallback; zCallback = 0;}
  };
};

/*
** This routine is invoked to handle a TCL trace callback.  The
** user data is a pointer to a TraceCallbackData structure
** which holds a pointer to the TCL interpreter and the name of
** a procedure in that interpreter.  This routine appends arguments
** to the TCL procedure and invokes it.  The arguments are:
**
**   1.  the name of the XML node that holds the trace
**   2.  the name of the XML node that changed
**   3.  the type of change ("create", "modify", or "delete")
**
** If the TCL command returns an error, then the error is reported
** as a background error and the trace is deleted.  
** See also XmlNode::invokeTrace().
*/
static int traceCallback(
  XmlCursor *pCursor,      // Ptr to node that changed
  XmlCursor *pReporter,    // Ptr to node holding the trigger
  void *pUserData,         // User specified callback data
  int chngType             // Type of change
){
  TraceCallbackData *pData = (TraceCallbackData*)pUserData;
  const char *zChngType;
  switch( chngType ){
    case XML_TRACE_CREATE:       zChngType = "create";   break;
    case XML_TRACE_MODIFY:       zChngType = "modify";   break;
    case XML_TRACE_DELETE:       zChngType = "delete";   break;
    default:                     zChngType = "unknown";  break;
  }

  /* Check for nodes that are marked for deletion.
  ** If the reporter is marked for deletion, it's a more serious error
  ** than if the modified node is:  in the former case we should delete
  ** the trace.
  */
  const char *zReportFull = pReporter->fullname();
  if( !zReportFull || !*zReportFull ){
    return 1;
  }
  const char *zCursorFull = pCursor->fullname();
  if( !zCursorFull || !*zCursorFull ){
    return 0;
  }

  assert(pData!=NULL);

  /* The callback may cause pData to change.  Save the interp so we
  ** can reference it later.
  */
  Tcl_Interp* interp = pData->interp;

  int rc = Tcl_VarEval(
    pData->interp,
    pData->zCallback,
    " ",
    pReporter->fullname(),
    " ",
    pCursor->fullname(),
    " ",
    zChngType,
    (char*)0
  );
  if( rc!=TCL_OK ){
    Tcl_BackgroundError(interp);
    return 1;
  }
  Tcl_ResetResult(interp);
  return 0;
}

/*
** This routine gets called whenever a trace is destroyed.  The
** user data for the trace being destroyed is passed in as the argument
** The role of this routine is to delete the user data.
*/
static void traceDeleteProc(void *pArg){
  TraceCallbackData *pData = (TraceCallbackData*)pArg;
  delete pData; pData = 0;
}

/*
** The following class implements a link between an XML element and
** a TCL variable.  The link keeps the two in sync at all times.
** whenever the XML is changed by an outside agent, the TCL is
** changed by this class to match the new XML value.  Similarly
** whenever the TCL variable changes, this class modifies the XML
** element to match.
*/
class XmlToTclLink {
private:
  Tcl_Interp *interp;  // The interpreter holding the linked variable
  char *zVarName;      // Name of the linked variable
  XmlCursor cur;       // Cursor on the XML element
  unsigned char busy;  // Blocks an infinite loop
  unsigned char dead;  // True if the XML trace has been deleted already

public:
  static int xmlChange(XmlCursor*, XmlCursor*, void*, int);
  static char *tclChange(void*, Tcl_Interp*, const char*, const char*, int);
  static void deleteProc(void*);
  const char *varname() const {return zVarName;};
  XmlToTclLink(Tcl_Interp *pInterp, const char *zName, XmlCursor *pCur);
  ~XmlToTclLink();
};

/*
** This is the constructor for a XML/TCL linkage.  The constructor
** sets up traces on both the XML and the TCL variable.  These traces
** invoke either xmlChange() or tclChange() when the XML or TCL
** changes, respectively.  The constructor also makes sure the TCL
** variable exists (creating it if necessary) and sets the value
** of the TCL variable to agree with the XML.  
*/
XmlToTclLink::XmlToTclLink(
  Tcl_Interp *pInterp,    // TCL Interpreter containing variable to link
  const char *zName,      // Name of the TCL variable to link
  XmlCursor *pCur         // Pointer to the XML element to link 
){
  interp = pInterp;
  zVarName = new char[strlen(zName)+1];
  cur = *pCur;
  busy = 0;
  dead = 0;
  if( !zVarName ) return;
  strcpy(zVarName, zName);
  xmlChange(pCur, pCur, this, XML_TRACE_MODIFY);
  Tcl_TraceVar(interp, zVarName, TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
      (Tcl_VarTraceProc*)tclChange, this);
  cur.trace(xmlChange, this, XML_TRACE_SELF, deleteProc);
}

/*
** The destructor for an XML/TCL linkage
*/
XmlToTclLink::~XmlToTclLink(){
  if( zVarName ){
    Tcl_UntraceVar(interp, zVarName, TCL_TRACE_WRITES|TCL_TRACE_UNSETS, 
       (Tcl_VarTraceProc*)XmlToTclLink::tclChange, this);
    delete [] zVarName; zVarName = 0;
  }
  if( !dead ){
    dead = 1;
    cur.untrace(xmlChange, this, XML_TRACE_SELF);
  }
};

/*
** This routine gets called when an XML element is deleted and was part
** of a XML/TCL linkage.  The job of this routine is to delete the
** link.
*/
void XmlToTclLink::deleteProc(void *pArg){
  XmlToTclLink *pData = (XmlToTclLink*)pArg;
  if( !pData->dead ){
    delete pData; pData = 0;
  }
}

/*
** This routine is invoked when the XML element of an XML/TCL linkage
** changes.  The job of this routine is to change the TCL variable
** to match the new XML value.
*/
int XmlToTclLink::xmlChange(
  XmlCursor *pCursor,      // Ptr to node that changed
  XmlCursor *pReporter,    // Ptr to node holding the trigger
  void *pUserData,         // User specified callback data
  int chngType             // Type of change
){
  XmlToTclLink *pData = (XmlToTclLink*)pUserData;
  if( pData->busy ) return 0;
  if( chngType==XML_TRACE_MODIFY ){
    const char *zXml = pCursor->get();
    int nXml = pCursor->length();
    Tcl_Obj *pObj;
    pObj = Tcl_GetVar2Ex(pData->interp, pData->zVarName, 0, TCL_GLOBAL_ONLY);
    if( pObj ){
      int nTcl;
      const char *zTcl = Tcl_GetStringFromObj(pObj, &nTcl);
      if( zTcl && zXml && (nTcl!=nXml || memcmp(zXml, zTcl, nTcl)!=0) ){
        pData->busy = 1;
        pObj = Tcl_NewStringObj(zXml, nXml);
        Tcl_SetVar2Ex(pData->interp, pData->zVarName, 0, pObj, TCL_GLOBAL_ONLY);
        pData->busy = 0;
      }
    }
  }
  return 0;
}

/*
** This routine is invoked by TCL when the value of a TCL variable
** changes and that TCL variable is linked to an XML element.
** This callback routine will turn around and change the value
** of the XML element.
*/
char *XmlToTclLink::tclChange(
  void *pUserData,
  Tcl_Interp *interp,
  const char *name1,
  const char *name2,
  int flags
){
  XmlToTclLink *pData = (XmlToTclLink*)pUserData;
  if( flags & TCL_TRACE_UNSETS ){
    delete pData; pData = 0;
  }else if( pData->busy==0 ){
    Tcl_Obj *pObj;
    const char *zTcl;
    int nTcl;
    pObj = Tcl_GetVar2Ex(interp, (char*)name1, name2, TCL_GLOBAL_ONLY);
    if( pObj ){
      pData->busy = 1;
      zTcl = (char*)Tcl_GetByteArrayFromObj(pObj, &nTcl);
      pData->cur.set(zTcl, nTcl);
      pData->busy = 0;
    }
  }
  return 0;
}

/*
** Given the command-line arguments shown, parse out the parameters
** of a trace.
**
** A trace command-line looks like this:
**
**      ?switches? ?PATH? PROC
**
** The option PATH may occur before or after the switches.  Valid
** switches are:  "-self", "-children", "-all", and "--".  Any
** number of switches can occur in any order.  The "--" switch means
** there are no more switches and any subsequent argument is either
** the PATH or the PROC.  There can be only one PATH.
*/
static int parseTraceCommandLine(
  Tcl_Interp *interp,      // The TCL interpreter
  int objc,                // Number of command-line arguments
  Tcl_Obj *const* objv,    // Values of command-line arguments
  XmlCursor *pCursor,      // Position this cursor according to PATH
  int *pFlags,             // Set this flag according to switches
  char **pzProc            // Write the PROC string here.
){
  int flags = 0;
  int no_more_sw = 0;
  int seen_path = 0;
  int i;
  for(i=0; i<objc-1; i++){
    char *z = Tcl_GetStringFromObj(objv[i], 0);
    if( z[0]=='-' && !no_more_sw ){
      if( strcmp(z,"-self")==0 ){
        flags |= XML_TRACE_SELF;
      }else if( strcmp(z,"-children")==0 ){
        flags |= XML_TRACE_CHILDREN;
      }else if( strcmp(z,"-all")==0 ){
        flags |= XML_TRACE_ALL;
      }else if( strcmp(z,"--")==0 ){
        no_more_sw = 1;
      }else{
        Tcl_AppendResult(interp,"unknown option: \"", z, "\"", 0);
        return TCL_ERROR;
      }
    }else if( seen_path ){
      Tcl_AppendResult(interp, "too many PATH parameters", 0);
      return TCL_ERROR;
    }else{
      pCursor->move(z);
      if( !pCursor->ok() ){
        Tcl_AppendResult(interp, "element \"", z, "\" does not exist", 0);
        return TCL_ERROR;
      }
      seen_path = 1;
    }
  }
  *pzProc = Tcl_GetStringFromObj(objv[objc-1], 0);
  if( flags==0 ){ flags = XML_TRACE_SELF; }
  *pFlags = flags;
  return TCL_OK;
}

/*
** Return the last N elements of the full pathname for pCur as
** a TCL object.  When N==1 the string returned is the equivalent
** of pCur->basename() (converted into a Tcl_Obj).  When N==2, the
** string is equivalent to "pCur->parent()->basename()/pCur->basename()".
** And so forth.
*/
static Tcl_Obj *RelativePath(int depth, XmlCursor *pCur){
  const char *zPath;
  int i;
  if( depth==0 ){
    return Tcl_NewStringObj("", -1);
  }
  zPath = pCur->fullname();
  for(i=strlen(zPath)-1; i>0; i--){
    if( zPath[i]=='/' ){
      depth--;
      if( depth<=0 ) break;
    }
  }
  return Tcl_NewStringObj(&zPath[i+1], -1);
}

/*
** A single member of this class contains a list of Xml cursors.
**
** In order to use this class correctly, after the list is created,
** use append() to append XML cursors to the list.
** When all of the paths have been appended to the list, call done().
** Then don't append any more XML cursors to it, 
** or modify it in any way, unless you call done() again.
**
** The get() method is used to return the i-th cursor in the list.
**
** This class is used to implement the "xml diff ... -apply" command.
*/
class XmlpList {
private:
  Tcl_Obj *pList;     // the Tcl list of XmlCursors
  XmlCursor **ppArgv; // the elements on the list, as XML cursor pointers
  int nobjc;          // the number of elements on the Tcl list
public:
  XmlpList()  {pList=Tcl_NewObj(); ppArgv=0; nobjc=0;}
  ~XmlpList() {
    Tcl_DecrRefCount(pList);
    for( int i=0; i<nobjc; ++i ){
      delete ppArgv[i]; ppArgv[i] = 0;
    }
    delete [] ppArgv; ppArgv = 0;
  }
  void append(const XmlCursor &cur){
    XmlCursor *pNew = new XmlCursor(cur);
    Tcl_ListObjAppendElement(0, pList, Tcl_NewWideIntObj((intptr_t)pNew));
  }
  void done() {
    if( ppArgv ){ delete [] ppArgv; ppArgv = 0; } 
    Tcl_Obj **ppObjv;
    Tcl_ListObjGetElements(0, pList, &nobjc, &ppObjv);
    ppArgv = new XmlCursor *[nobjc];
    for( int i=0; i<nobjc; ++i ){
      Tcl_GetWideIntFromObj(0, ppObjv[i], (Tcl_WideInt *)&ppArgv[i]);
    }
  }
  int count() const {return nobjc;}
  XmlCursor *get(int i) const {return ppArgv[i];}
};

/*
** Delete the XML relative path contained in the Tcl object 
** pointed to by pPathObj from each of the XML root paths on the
** list pointed to by pRootList.
*/
static void DeleteFromRoots(
  XmlpList *pRootList, // list of XML cursors pointing to XML paths
                       // relative to which the deletes should be applied
  Tcl_Obj *pPathObj
){
  char *zRelpath = Tcl_GetStringFromObj(pPathObj, 0);
  for( int i=0; i<pRootList->count(); ++i ){
    XmlCursor *pCur = pRootList->get(i);
    if( pCur ){
      pCur->unlink(zRelpath);
    }
  }
}

/*
** Set the value pointed to by pValue to a set of XML paths.
** pRootList contains the names of all of the XML path roots used to
** generate the XML paths.  pRelpath contains the relative path to use within
** each of these roots.
*/
static void SetValueInRoots(
  XmlpList *pRootList, // list of XML cursors pointing to XML paths
                       // relative to which the sets should be applied
  Tcl_Obj *pRelpath,
  Tcl_Obj *pValue
){
  int nValue;
  char *zRelpath = Tcl_GetStringFromObj(pRelpath, 0);
  char *zValue = (char*)Tcl_GetByteArrayFromObj(pValue, &nValue);
  for( int i=0; i<pRootList->count(); ++i ){
    XmlCursor *pCur = pRootList->get(i);
    if( pCur ){
      const char *zRoot = pCur->fullname();
      pCur->move(zRelpath);
      if( !pCur->ok() ){
         if( !pCur->create(zRelpath) ) return;  // trace proc killed create
      }
      pCur->set(zValue, nValue);
      pCur->move(zRoot);
    }
  }
}

/*
** Delete the attribute given at a set of XML paths.
** pRootList contains the names of all of the XML path roots used to
** generate the XML paths.  pPathObj contains the relative path to use within
** each of these roots.
*/
static void DeleteAttrFromRoots(
  XmlpList *pRootList, // List of XML cursors pointing to XML paths
                       // relative to which the deletes should be applied
  Tcl_Obj *pPathObj,   // Relative path to node at which attr deletes occur
  const char *zName    // Name of the attribute to be deleted
){
  char *zRelpath = Tcl_GetStringFromObj(pPathObj, 0);
  for( int i=0; i<pRootList->count(); ++i ){
    XmlCursor *pCur = pRootList->get(i);
    if( pCur ){
      XmlCursor cur2 = *pCur;
      cur2.move(zRelpath);
      cur2.removeAttribute(zName);
    }
  }
}

/*
** Set the value of the attribute given at a set of XML paths.
** pRootList contains the names of all of the XML path roots used to
** generate the XML paths.  pRelpath contains the relative path to use within
** each of these roots.
*/
static void SetAttrInRoots(
  XmlpList *pRootList, // list of XML cursors pointing to XML paths
                       // relative to which the sets should be applied
  Tcl_Obj *pRelpath,
  const char *zName,
  const char *zValue
){
  char *zRelpath = Tcl_GetStringFromObj(pRelpath, 0);
  for( int i=0; i<pRootList->count(); ++i ){
    XmlCursor *pCur = pRootList->get(i);
    if( pCur ){
      XmlCursor cur2 = *pCur;
      cur2.move(zRelpath);
      if( !cur2.ok() ){
         if( !cur2.create(zRelpath) ) return;  // Create failed due to trace
      }
      cur2.addAttribute(zName, zValue);
    }
  }
}

/*
** If pEditList is nonzero, generate "attr set" edit instructions 
** for all attributes at the node under the given cursor.  
**
** If useNew==1, then use the keyword "new" instead of "set".
**
** If pRootList is nonzero, set the attributes in each tree rooted by the
** XML paths in pRootList.
*/
static void AttrSet(
  int depth, 
  Tcl_Obj *pEditList,  // If nonzero, write the edit script here
  XmlpList *pRootList, // If nonzero, contains a list of XML cursors pointing 
                       // to root nodes.  The edits should be applied relative
                       // to all of these nodes.
  XmlCursor *pCur,     // The tree that needs to be changed
  int useNew
){
  Tcl_Obj *pRelpath = RelativePath(depth, pCur);
  XmlCursor cur2 = *pCur;
  Tcl_IncrRefCount(pRelpath);
  for(cur2.first_attr(); cur2.attr_ok(); cur2.next_attr()){
    if( pEditList ){
      Tcl_ListObjAppendElement(0, pEditList, Tcl_NewStringObj("attr",4));
      Tcl_ListObjAppendElement(0, pEditList,
                               Tcl_NewStringObj(useNew?"new":"set",3));
      Tcl_ListObjAppendElement(0, pEditList, pRelpath);
      Tcl_ListObjAppendElement(0, pEditList,
                               Tcl_NewStringObj(cur2.getAttrName(),
                               strlen(cur2.getAttrName())));
      Tcl_ListObjAppendElement(0, pEditList,
                               Tcl_NewStringObj(cur2.getAttribute(),
                               strlen(cur2.getAttribute())));
    }
    Tcl_IncrRefCount(pRelpath);
    if( pRootList ){
      SetAttrInRoots(pRootList, pRelpath, cur2.getAttrName(), 
                     cur2.getAttribute());
    }
    Tcl_DecrRefCount(pRelpath);
  }
  Tcl_DecrRefCount(pRelpath);
}

/*
** If pEditList is nonzero, generate "set" edit instructions 
** for all leaves under the given cursor.  
** If pCur is itself a leaf, then just one
** "set" instruction is added to pList.  If pCur has subelements,
** then a separate "set" instruction is added for each subelement.
**
** If useNew==1, then use the keyword "new" instead of "set".
**
** If pRootList is nonzero, set the value in each tree rooted by the
** XML paths in pRootList.
*/
static void DeepSet(
  int depth, 
  Tcl_Obj *pEditList,  // If nonzero, write the edit script here
  XmlpList *pRootList, // If nonzero, contains a list of XML cursors pointing 
                       // to root nodes.  The edits should be applied relative
                       // to all of these nodes.
  XmlCursor *pCur,     // The tree that needs to be changed
  int useNew
){
  if( pCur->is_leaf() ){
    Tcl_Obj *pValue;
    if( pCur->is_real() ){
      if( pCur->is_int() ){
        pValue = Tcl_NewIntObj((int)pCur->get_real());
      }else{
        pValue = Tcl_NewDoubleObj(pCur->get_real());
      }
    }else{
      pValue = Tcl_NewByteArrayObj((unsigned char*)pCur->get(), pCur->length());
    }
    Tcl_Obj *pRelativePath = RelativePath(depth, pCur);
    if( pEditList ){
      Tcl_ListObjAppendElement(0, pEditList, 
          Tcl_NewStringObj(useNew?"new":"set",3));
      Tcl_ListObjAppendElement(0, pEditList, pRelativePath);
      Tcl_ListObjAppendElement(0, pEditList, pValue);
    }
    Tcl_IncrRefCount(pRelativePath);
    Tcl_IncrRefCount(pValue);
    if( pRootList ){
      SetValueInRoots(pRootList, pRelativePath, pValue);
    }
    Tcl_DecrRefCount(pValue);
    Tcl_DecrRefCount(pRelativePath);
  }else{
    XmlCursor cur = *pCur;
    for(cur.move_first(); cur.ok(); cur.move_next()){
      DeepSet(depth+1, pEditList, pRootList, &cur, useNew);
    }
  }
  AttrSet(depth, pEditList, pRootList, pCur, useNew);
}

static void AttrDiff(
  int depth,           // Depth of the recursion. Needed for relative pathnames
  Tcl_Obj *pEditList,  // If nonzero, write the edit script here
  XmlpList *pRootList, // If nonzero, contains a list of XML cursors pointing
                       // to root nodes.  The edits should be applied relative
                       // to all of these nodes.
  XmlCursor *pTarget,  // The tree that pRef should be made to look like
  XmlCursor *pRef,     // The tree that needs to be changed
  int useNew           // Use "new" for elements missing from pRef
){
  /*
  ** First figure out what attributes are in pRef but missing from pTarget.
  ** These attributes will need to be deleted from pRef.
  */
  XmlCursor ref = *pRef;
  Tcl_Obj *pRelpath = RelativePath(depth, &ref);
  Tcl_IncrRefCount(pRelpath);
  ref.first_attr();
  while( ref.attr_ok() ){
    if ( !pTarget->attr_exists(ref.getAttrName()) ){
      if( pEditList ){
        Tcl_ListObjAppendElement(0, pEditList, Tcl_NewStringObj("attr",4));
        Tcl_ListObjAppendElement(0, pEditList, Tcl_NewStringObj("unset",5));
        Tcl_ListObjAppendElement(0, pEditList, pRelpath);
        Tcl_ListObjAppendElement(0, pEditList,
                Tcl_NewStringObj(ref.getAttrName(),strlen(ref.getAttrName())));
      }
      Tcl_IncrRefCount(pRelpath);
      if( pRootList ){
        DeleteAttrFromRoots(pRootList, pRelpath, ref.getAttrName());
      }
      Tcl_DecrRefCount(pRelpath);
      ref.next_attr();
    }else{
      ref.next_attr();
    }
  }
  
  /*
  ** Now find attributes of pTarget that are either missing in pRef or have
  ** a different value in pRef.  These attributes need to be added or changed
  ** in pRef.
  */
  XmlCursor target = *pTarget;
  for(target.first_attr(); target.attr_ok(); target.next_attr()){
    if( ref.attr_exists(target.getAttrName()) ){
      if( strcmp(target.getAttribute(),
                 ref.getAttribute(target.getAttrName())) ){
        if( pEditList ){
          Tcl_ListObjAppendElement(0, pEditList, Tcl_NewStringObj("attr",4));
          Tcl_ListObjAppendElement(0, pEditList, Tcl_NewStringObj("set",3));
          Tcl_ListObjAppendElement(0, pEditList, pRelpath);
          Tcl_ListObjAppendElement(0, pEditList,
                                   Tcl_NewStringObj(target.getAttrName(),
                                   strlen(target.getAttrName())));
          Tcl_ListObjAppendElement(0, pEditList,
                                   Tcl_NewStringObj(target.getAttribute(),
                                   strlen(target.getAttribute())));
        }
        Tcl_IncrRefCount(pRelpath);
        if( pRootList ){
          SetAttrInRoots(pRootList, pRelpath, target.getAttrName(), 
                         target.getAttribute());
        }
        Tcl_DecrRefCount(pRelpath);
      }
    }else{
      if( pEditList ){
        Tcl_ListObjAppendElement(0, pEditList, Tcl_NewStringObj("attr",4));
        Tcl_ListObjAppendElement(0, pEditList,
                                 Tcl_NewStringObj(useNew?"new":"set",3));
        Tcl_ListObjAppendElement(0, pEditList, pRelpath);
        Tcl_ListObjAppendElement(0, pEditList,
                                 Tcl_NewStringObj(target.getAttrName(),
                                 strlen(target.getAttrName())));
        Tcl_ListObjAppendElement(0, pEditList,
                                 Tcl_NewStringObj(target.getAttribute(),
                                 strlen(target.getAttribute())));
      }
      Tcl_IncrRefCount(pRelpath);
      if( pRootList ){
        SetAttrInRoots(pRootList, pRelpath, target.getAttrName(), 
                       target.getAttribute());
      }
      Tcl_DecrRefCount(pRelpath);
    }
  }
  Tcl_DecrRefCount(pRelpath);
}

/*
** Generate a list of edit instructions that can be used to convert the
** XML tree at *pRef into a duplication of the tree at *pTarget,
** if pEditList is nonzero.
**
** Apply the edits into a list of XML paths on the list
** pointed to by pRootList, if pRootList is nonzero.
**
** Edit instructions take one of the following forms:
** set RELPATH VALUE
** new RELPATH VALUE
** unset RELPATH
** attr set RELPATH ATTRNAME VALUE
** attr new RELPATH ATTRNAME VALUE
** attr unset RELPATH ATTRNAME
*/
static void DeepDiff(
  int depth,           // Depth of the recursion. Needed for relative pathnames
  Tcl_Obj *pEditList,  // If nonzero, write the edit script here
  XmlpList *pRootList, // If nonzero, contains a list of XML cursors pointing 
                       // to root nodes.  The edits should be applied relative
                       // to all of these nodes.
  XmlCursor *pTarget,  // The tree that pRef should be made to look like
  XmlCursor *pRef,     // The tree that needs to be changed
  int useNewKeyword    // Use "new" for elements missing from pRef
){
  /*
  ** First figure out what elements are in pRef but missing from pTarget.
  ** These elements will need to be deleted from pRef.
  */
  XmlCursor ref = *pRef;
  ref.move_first();
  while( ref.ok() ){
    if( !pTarget->exists(ref.basename()) ){
      Tcl_Obj *pRelpath = RelativePath(depth, &ref);
      if( pEditList ){
        Tcl_ListObjAppendElement(0, pEditList, Tcl_NewStringObj("unset",5));
        Tcl_ListObjAppendElement(0, pEditList, pRelpath);
      }
      ref.move_next();
      Tcl_IncrRefCount(pRelpath);
      if( pRootList ){
        DeleteFromRoots(pRootList, pRelpath);
      }
      Tcl_DecrRefCount(pRelpath);
    }else{
      ref.move_next();
    }
  }

  /*
  ** If pTarget is a leaf, then pRef will have to be a leaf with the same value.
  */
  ref = *pRef;
  XmlCursor target = *pTarget;
  if( target.is_leaf() ){
    if( ref.is_leaf() ){
      int same = ref.equal(ref, target);
      if( !same ){
        DeepSet(depth-1, pEditList, pRootList, &target, 0);
      }else{
        AttrDiff(depth-1, pEditList, pRootList, &target, &ref, useNewKeyword);
      }
    }else{
      if( target.get()!=0 && target.length()!=0 ){
        DeepSet(depth-1, pEditList, pRootList, &target, 0);
      }
    }
    return;
  }

  /*
  ** Now find elements of pTarget that are either missing in pRef or have
  ** a different value in pRef.  These elements need to be added or changed
  ** in pRef.
  */
  for(target.move_first(); target.ok(); target.move_next()){
    if( ref.move(target.basename()) ){
      if( ref.is_leaf() && target.is_leaf() ){
        int same = ref.equal(ref, target);
        if( !same ){
          DeepSet(depth, pEditList, pRootList, &target, 0);
        }else{
          AttrDiff(depth, pEditList, pRootList, &target, &ref, useNewKeyword);
        }
      }else{
        DeepDiff(depth+1, pEditList, pRootList, &target, &ref, useNewKeyword);
        AttrDiff(depth, pEditList, pRootList, &target, &ref, useNewKeyword);
      }
      ref.move_up();
    }else{
      DeepSet(depth, pEditList, pRootList, &target, useNewKeyword);
    }
  }
}

/*
** This routine runs when the TCL interpreter executes one of the
** commands associated with an Xml cursor.  The first argument to
** the TCL command is the name of a method to call for the cursor.
** This routine figures out which method was invoked and takes
** appropriate action.
*/
static int XmlCmdProc(
  ClientData pData,       // A pointer to an XmlCursor
  Tcl_Interp *interp,      // The interpreter controlling the new cursor
  int objc,                // Number of arguments
  Tcl_Obj *const* objv     // Value of each argument
){
  static const char *XML_strs[] = {
    "align",              "attr",              "basename",
    "boolean",            "cat",               "cd",
    "check",              "close",             "compare",
    "copy",               "count",             "create",
    "diff",               "directory",         "exists",
    "first",              "fullname",          "get",
    "hash",               "index",             "isdir",
    "isleaf",             "isreal",            "link",
    "list",               "ls",                "memused",
    "mkdir",              "mv",                "name",
    "next",               "ok",                "parse",
    "pwd",                "realvalue",         "render",
    "renumber",           "reset_counter",     "rmdir",
    "set",                "sync",              "trace",
    "unlink",             "unset",             "untrace",
    0
  };
  enum XML_enum {
    XML_ALIGN,            XML_ATTR,            XML_BASENAME,
    XML_BOOLEAN,          XML_CAT,             XML_CD,
    XML_CHECK,            XML_CLOSE,           XML_COMPARE,
    XML_COPY,             XML_COUNT,           XML_CREATE,
    XML_DIFF,             XML_DIRECTORY,       XML_EXISTS,
    XML_FIRST,            XML_FULLNAME,        XML_GET,
    XML_HASH,             XML_INDEX,           XML_ISDIR,
    XML_ISLEAF,           XML_ISREAL,          XML_LINK,
    XML_LIST,             XML_LS,              XML_MEMUSED,
    XML_MKDIR,            XML_MV,              XML_NAME,
    XML_NEXT,             XML_OK,              XML_PARSE,
    XML_PWD,              XML_REALVALUE,       XML_RENDER,
    XML_RENUMBER,         XML_RESET_COUNTER,   XML_RMDIR,
    XML_SET,              XML_SYNC,            XML_TRACE,
    XML_UNLINK,           XML_UNSET,           XML_UNTRACE,
  };
  XmlCursor *pCur = (XmlCursor*)pData;
  int index;

  if( objc<2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "METHOD ?ARG ...?");
    return TCL_ERROR;
  }
  if( Tcl_GetIndexFromObj(interp, objv[1], XML_strs, "option", 0, &index)){
    return TCL_ERROR;
  }
  switch( (enum XML_enum)index ){
    /*
    ** xmlp-tcl-cmd: CURSOR align ?PATH? TARGET
    **
    ** Renumber the indices of all descendents of TARGET so that they match
    ** (as much as possible) the indices on descendents of PATH.  If PATH
    ** is omitted, use "." as a default.
    */
    case XML_ALIGN: {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH? TARGET");
        return TCL_ERROR;
      }
      XmlCursor basis, target;
      basis = *pCur;
      target = *pCur;
      if( objc==3 ){
        target.move(Tcl_GetStringFromObj(objv[2],0));
      }else{
        basis.move(Tcl_GetStringFromObj(objv[2],0));
        target.move(Tcl_GetStringFromObj(objv[3],0));
      }
      basis.align(&target);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR attr SUBCOMMAND ...
    **
    ** Execute one of various subcommands used to manipulate attributes
    */
    case XML_ATTR: {
      if( objc<3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "SUBCOMMAND ?ARG ...?");
        return TCL_ERROR;
      }
      return AttrCmdProc(interp, objc, objv, pCur);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR basename ?PATH?
    **
    ** Return the basename of the XML node at which the cursor is currently
    ** pointing.  The basename consists of the name of the current node
    ** followed by a colon and the index number of the current node.  The
    ** names of parent nodes are omitted.  Example: "event:4".
    */
    case XML_BASENAME: {
      if( objc!=2 && objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH?");
        return TCL_ERROR;
      }
      XmlCursor cur = *pCur;
      if( objc==3 ){
        const char *zPath = Tcl_GetStringFromObj(objv[2], 0);
        cur.move(zPath);
      }
      Tcl_AppendResult(interp, cur.basename(), 0);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR boolean PATH ?VALUE?
    **
    ** Without a VALUE argument, this method returns true or false.  It returns
    ** true if PATH exists and is not zero.  It returns false if PATH does not
    ** exist or has a value of zero.
    **
    ** If the argument VALUE exists and is true, then PATH is set to 1.
    ** If the argument VALUE exists and is false, then PATH is deleted.
    */
    case XML_BOOLEAN: {
      if( objc!=3 && objc!=4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "PATH ?VALUE?");
        return TCL_ERROR;
      }
      XmlCursor cur = *pCur;
      const char *zPath = Tcl_GetStringFromObj(objv[2], 0);
      if( objc==3 ){
        cur.move(zPath);
        int r;
        if( !cur.ok() ){
          r = 0;
        }else if( cur.is_real() ){
          r = cur.get_real()!=0.0;
        }else{
          const char *z = cur.get();
          r = z!=0 && z[0]!=0 && z[0]!='0';
        }
        Tcl_SetObjResult(interp, Tcl_NewBooleanObj(r));
      }else{
        int v;
        if( Tcl_GetBooleanFromObj(interp, objv[3], &v) ) return TCL_ERROR;
        cur.set_boolean(zPath, v);
      }
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR cd ?-create? PATH
    **
    ** Change the cursor to point to the XML node identified by PATH.
    ** If PATH does not exist and the -create option is used then the
    ** node is created.  If PATH does not exist and the -create option
    ** is omitted, then the cursor may be left unchanged or it may be
    ** left pointing to some parent of PATH that does happen to exist.
    */
    case XML_CD: {
      if( objc!=3 && objc!=4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?-create? PATH");
        return TCL_ERROR;
      }
      if( objc==4 ){
        char *zOpt = Tcl_GetStringFromObj(objv[2],0);
        if( strcmp(zOpt,"-create")!=0 ){
          Tcl_AppendResult(interp,"unknown option: ", zOpt, 0);
          return TCL_ERROR;
        }
        pCur->moveOrCreate(Tcl_GetStringFromObj(objv[3],0));  // may fail
      }else{
        pCur->move(Tcl_GetStringFromObj(objv[2],0));
      }
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, pCur->fullname(), NULL);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR close
    **
    ** Close the cursor.  This is the same as deleting the cursor command.
    */
    case XML_CLOSE: {
      Tcl_DeleteCommand(interp, Tcl_GetStringFromObj(objv[0], 0));
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR check TEXT
    **
    ** Do syntax checking to verify that the XML contained in the TEXT argument
    ** is well-formed.  If the XML is well-formed, then return an empty string.
    ** If the XML is not well-formed, then this method throws an exception
    ** that describes the error.
    */
    case XML_CHECK: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "XML");
        return TCL_ERROR;
      }
      int nText;
      char *zText = Tcl_GetStringFromObj(objv[2], &nText);
      const char *zErrMsg;
      int iOffset, iLength;
      int rc = XmlCursor::check(zText, nText, &zErrMsg, &iOffset, &iLength);
      if( rc ){
        Tcl_DString str;
        Tcl_DStringInit(&str);
        Tcl_DStringAppend(&str, zErrMsg, -1);
        Tcl_DStringAppend(&str, ": ", 2);
        Tcl_DStringAppend(&str, &zText[iOffset], iLength);
        Tcl_DStringResult(interp, &str);
        return TCL_ERROR;
      }
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR compare ?SRC? DEST
    **
    ** Return TRUE if there are any differences between SRC and DEST.
    ** Return FALSE if SRC and DEST are identical.
    ** If the SRC argument is omitted then the node at which the
    ** cursor is pointing is used instead.
    */
    case XML_COMPARE: {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?SRC? DEST");
        return TCL_ERROR;
      }
      XmlCursor from, to;
      from = *pCur;
      to = *pCur;
      if( objc==3 ){
        to.move(Tcl_GetStringFromObj(objv[2],0));
      }else{
        from.move(Tcl_GetStringFromObj(objv[2],0));
        to.move(Tcl_GetStringFromObj(objv[3],0));
      }
      if( !from.ok() && !to.ok() ){
        Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
      }else if( !from.ok() || !to.ok() ){
        Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 1);
      }else{
        Tcl_SetBooleanObj(Tcl_GetObjResult(interp), from.Compare(&to));
      }
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR copy ?SRC? DEST
    **
    ** Make a copy of the XML tree rooted at SRC in a different location
    ** specified by DEST.  If the SRC argument is omitted then a copy is
    ** made of the node at which the cursor is pointing.  
    */
    case XML_COPY: {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?SRC? DEST");
        return TCL_ERROR;
      }
      XmlCursor from, to;
      from = *pCur;
      to = *pCur;
      if( objc==3 ){
        if( !to.moveOrCreate(Tcl_GetStringFromObj(objv[2],0)) ) break;
      }else{
        if( !from.move(Tcl_GetStringFromObj(objv[2],0)) ) break;
        if( !to.moveOrCreate(Tcl_GetStringFromObj(objv[3],0)) ) break;
      }
      if( strcmp(from.fullname(), "/")==0 ){
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "Can't copy from the root.", NULL);
        return TCL_ERROR;
      }
      if( from.is_ancestor(&to) ){
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "Can't copy ", from.fullname(),
            " to its descendent ", to.fullname(), ".", NULL);
        return TCL_ERROR;
      }
      if( from!=to ){
        from.DeepCopy(&from, &to);
      }
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR count PATTERN
    **
    ** For the XML node at which the cursor is pointing, return the number
    ** of child elements whose names match the given pattern.  
    */
    case XML_COUNT: {
      XmlCursor cur2 = *pCur;
      int cnt = 0;
      char *zPattern;
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "PATTERN");
        return TCL_ERROR;
      }
      zPattern = Tcl_GetStringFromObj(objv[2], 0);
      for(cur2.move_first(); cur2.ok(); cur2.move_next()){
        if( globCompare(cur2.name(), zPattern) ) cnt++;
      }
      Tcl_SetIntObj(Tcl_GetObjResult(interp), cnt);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR create NAME
    **
    ** Create a new XML node called NAME and move the cursor so that it
    ** points to the new node.  See also: mkdir
    */
    case XML_CREATE: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "PATH");
        return TCL_ERROR;
      }
      const char *z = pCur->create(Tcl_GetStringFromObj(objv[2], 0));
      if( z ) Tcl_AppendElement(interp, z);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR diff ?-new? ?SRC? DEST ?-apply NODE?...
    **
    ** Return a string that describes the differences between the XML
    ** subtrees at SRC and DEST.  If the SRC argument is omitted, use
    ** the node at which the cursor is pointing instead.
    **
    ** The string describes a set of actions that would be required to
    ** convert DEST into an exact copy of SRC.  Each edit instruction
    ** begins with a keyword "unset" or "set".  If the keyword is "unset"
    ** then the next token is the name of an XML element to remove from
    ** DEST.  If the keyword is "set" then the next two tokens are the
    ** name of an element in DEST to create or change and the new value
    ** for that element, respectively.
    **
    ** The default behavior is to generate a "set" instruction for elements
    ** that are different between SRC and DEST and for elements that are
    ** missing from SRC.  But if the "-new" option is added, elements that
    ** are missing from SRC are given a "new" tag in place of "set".  The
    ** -new option can be used to construct a 3-way diff.
    **
    ** If there are one or more "-apply NODE" arguments, then the result of
    ** the "diff" operation is applied to each tree rooted at NODE.
    */
    case XML_DIFF: {
      if( objc<3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?-new? ?SRC? DEST ?-apply NODE?...");
        return TCL_ERROR;
      }
      XmlCursor from = *pCur;
      XmlCursor to = from;
      int useNewKeyword = 0;
      if( objc>=4 && strcmp(Tcl_GetStringFromObj(objv[2],0),"-new")==0 ){
        useNewKeyword = 1;
        objc--;
        objv++;
      }
      if( (objc%2)==1 ){
        to.move(Tcl_GetStringFromObj(objv[2],0));
      }else{
        from.move(Tcl_GetStringFromObj(objv[2],0));
        to.move(Tcl_GetStringFromObj(objv[3],0));
        objc--;
        objv++;
      }
      if( objc>3 ){
        XmlpList rootList;
        XmlCursor cur;
        for( int i=3; i<objc; i+=2 ){
          if( strcmp(Tcl_GetStringFromObj(objv[i],0), "-apply")!=0 ){
            Tcl_WrongNumArgs(interp, 2, objv, 
                "?-new? ?SRC? DEST ?-apply NODE?...");
            return TCL_ERROR;
          }
          char *zRoot = Tcl_GetStringFromObj(objv[i+1], 0);
          cur.move(zRoot);
          if( !cur.ok() ){
            if( cur.create(zRoot) ){  // may fail due to trace proc
              rootList.append(cur);
            }
          }else{
            rootList.append(cur);
          }
        }
        rootList.done();
        Tcl_Obj *pEditList = Tcl_NewObj();
        DeepDiff(1, pEditList, &rootList, &from, &to, useNewKeyword);
        Tcl_SetObjResult(interp, pEditList);
      }else{
        Tcl_Obj *pEditList = Tcl_NewObj();
        DeepDiff(1, pEditList, 0, &from, &to, useNewKeyword);
        Tcl_SetObjResult(interp, pEditList);
      }
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR directory
    **
    ** Return the full pathname of the directory that contains the
    ** XML object to which the cursor is pointing.
    */
    case XML_DIRECTORY: {
      Tcl_AppendResult(interp, pCur->pwd(), 0);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR exists PATH
    **
    ** Return TRUE or FALSE depending on whether or not there exists
    ** an XML element named PATH.
    */
    case XML_EXISTS: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "PATH");
        return TCL_ERROR;
      }
      Tcl_SetIntObj(Tcl_GetObjResult(interp), 
          pCur->exists(Tcl_GetStringFromObj(objv[2],0)));
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR first ?PATTERN?
    **
    ** Move the cursor so that it points to the first child element
    ** that matches PATTERN.  If PATTERN is omitted, then just move
    ** the cursor to point to the first first child element.  This
    ** method can be used in combination with the "next" method to
    ** implement a loop over all children of an XML node.
    */
    case XML_FIRST: {
      XmlCursor cur2 = *pCur;
      int rc;
      pCur->move_first();
      if( objc>2 ){
        char *zPattern = Tcl_GetStringFromObj(objv[2],0);
        while( pCur->ok() && globCompare(pCur->name(), zPattern)==0 ){
          pCur->move_next();
        }
      }
      rc = pCur->ok();
      if( !rc ){ *pCur = cur2; }
      Tcl_SetIntObj(Tcl_GetObjResult(interp), rc);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR pwd ?PATH?
    **
    ** Return the full pathname of the object identified by PATH.  Or
    ** if PATH is omitted, return the full pathname of the object at which
    ** the cursor is pointing.
    */
    /*
    ** xmlp-tcl-cmd: CURSOR fullname ?PATH?
    **
    ** This command is an alias for the "pwd" command.  See the description
    ** of the "pwd" command for details.
    */
    case XML_FULLNAME:
    case XML_PWD: {
      if( objc!=2 && objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH?");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      if( objc==3 ){
        cur2.move(Tcl_GetStringFromObj(objv[2], 0));
        if( !cur2.ok() ) break;
      }
      Tcl_AppendResult(interp, cur2.fullname(), 0);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR get ?PATH? ?VALUE?
    **
    ** Return the content of the leaf XML node at PATH.  If PATH is
    ** omitted, then return the content of the node at which the cursor
    ** is pointing.  If the specified node is not a leaf, then an empty
    ** string is returned.  If the specified node does not exist, then
    ** VALUE is returned if it is specified and an empty string is returned
    ** if VALUE is omitted.
    */
    /*
    ** xmlp-tcl-cmd: CURSOR cat ?PATH? ?VALUE?
    **
    ** This is an alias for the "get" command.  See the documentation on
    ** the "get" command for details.
    */
    case XML_GET:
    case XML_CAT: {
      if( objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH? ?VALUE?");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      if( objc>=3 ){
        cur2.move(Tcl_GetStringFromObj(objv[2], 0));
        if( !cur2.ok() ){
          if( objc==4 ){
            Tcl_SetObjResult(interp,Tcl_DuplicateObj(objv[3]));
          }
          break;
        }
      }
      Tcl_Obj *pObj = Tcl_NewByteArrayObj((const unsigned char*)cur2.get(),
                          cur2.length());
      Tcl_SetObjResult(interp, pObj);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR index ?PATH?
    **
    ** Return the numeric index that appears at the end of the name
    ** for the XML node specified by PATH.  If PATH is omitted, use the 
    ** node at which the cursor is pointing.
    */
    case XML_INDEX: {
      if( objc>3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH?");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      if( objc==3 ){
        cur2.move(Tcl_GetStringFromObj(objv[2], 0));
        if( !cur2.ok() ) break;
      }
      Tcl_SetIntObj(Tcl_GetObjResult(interp), cur2.index());
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR hash ?ARG ARG...?
    **
    ** Hook into the hash methods.  See xmlphash.cpp for details.
    */
    case XML_HASH: {
      if( objc<3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "SUBCMD ?ARG ARG...?");
        return TCL_ERROR;
      }
      return HashProc(interp, objc, objv, pCur);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR isdir ?PATH?
    **
    ** Return TRUE if the XML node specified by PATH is not a leaf node.
    ** Return FALSE if it is a leaf node.  If PATH is omitted, use the
    ** node at which the cursor is pointing.
    */
    /*
    ** xmlp-tcl-cmd: CURSOR isleaf ?PATH?
    **
    ** Return TRUE if the XML node specified by PATH is a leaf node.
    ** Return FALSE if it is not a leaf node.  If PATH is omitted, use the
    ** node at which the cursor is pointing.
    */
    case XML_ISDIR:
    case XML_ISLEAF: {
      if( objc>3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH?");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      if( objc==3 ){
        cur2.move(Tcl_GetStringFromObj(objv[2], 0));
        if( !cur2.ok() ){
          Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
          break;
        }
      }
      int r = cur2.is_leaf();
      if( index==XML_ISDIR ) r = !r;
      Tcl_SetBooleanObj(Tcl_GetObjResult(interp), r);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR isreal ?PATH?
    **
    ** Return TRUE if the XML node specified by PATH contains a real value.
    ** Return FALSE if it is does not.  If PATH is omitted, use the
    ** node at which the cursor is pointing.
    */
    case XML_ISREAL: {
      if( objc>3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH?");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      if( objc==3 ){
        cur2.move(Tcl_GetStringFromObj(objv[2], 0));
        if( !cur2.ok() ){
          Tcl_SetBooleanObj(Tcl_GetObjResult(interp), 0);
          break;
        }
      }
      Tcl_SetBooleanObj(Tcl_GetObjResult(interp), cur2.is_real());
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR link ?-info? ?PATH? VARIABLE
    **
    ** Create a linkage between the XML node identified by PATH and the
    ** Tcl variable named VARIABLE.  If the PATH argument is omitted, use
    ** the XML node at which the cursor is pointing instead.  The XML node
    ** must be a leaf node.  Any changes made to the XML node are 
    ** automatically copied over the variable and any changes to the 
    ** variable are automatically copied into the XML node.  Deleting either
    ** the XML node or the variable breaks the link.
    **
    ** If the -info option is specified, then the return value is a list
    ** of all variables linked to the specified XML node.  When the
    ** -info option is used, the VARIABLE name should be omitted.
    */
    case XML_LINK: {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH? VARNAME");
        return TCL_ERROR;
      }
      XmlCursor cur = *pCur;
      if( strcmp(Tcl_GetStringFromObj(objv[2], 0),"-info")==0 ){
        if( objc==4 ){
          cur.move(Tcl_GetStringFromObj(objv[3], 0));
          if( !cur.ok() ) break;
        }
        XmlTraceCallback xFunc;
        XmlToTclLink *pUserData;
        int flags;
        for(int i=0; cur.getTrace(i,&xFunc,(void**)&pUserData,&flags); i++){
          if( xFunc!=XmlToTclLink::xmlChange ) continue;
          Tcl_AppendElement(interp, pUserData->varname());
        }
        break;
      }
      char *zVarName;
      if( objc==4 ){
        char *zPath = Tcl_GetStringFromObj(objv[2], 0);
        cur.move(zPath);
        if( !cur.ok() ){
          Tcl_AppendResult(interp, "no such XML element: ", zPath, 0);
          return TCL_ERROR;
        }
        zVarName = Tcl_GetStringFromObj(objv[3], 0);
      }else{
        zVarName = Tcl_GetStringFromObj(objv[2], 0);
      }
      new XmlToTclLink(interp, zVarName, &cur);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR ls ?PATH ?-noindex? ?
    **
    ** List all children of the XML node identified by PATH.  If PATH
    ** is omitted, then list all children of the node that the cursor
    ** is pointing to.
    **
    ** If -noindex is given, don't include the children's indices in
    ** the returned list.
    */
    /*
    ** xmlp-tcl-cmd: CURSOR list ?PATH ?-noindex? ?
    **
    ** This command is an alias for "ls".  See the description of the "ls"
    ** command for additional information.
    */
    case XML_LIST:
    case XML_LS: {
      Tcl_Obj *pList;
      XmlCursor cur2 = *pCur;
      int bNoIndex = 0;
      if( objc>=3 ){
        cur2.move(Tcl_GetStringFromObj(objv[2],0));
        if( !cur2.ok() ) break;
        if( objc>=4 ){
          if( strcmp(Tcl_GetStringFromObj(objv[3], 0), "-noindex")==0 ){
            bNoIndex = 1;
          }
        }
      }
      pList = Tcl_NewListObj(0, 0);
      for(cur2.move_first(); cur2.ok(); cur2.move_next()){
        Tcl_Obj *pElem;
        if( bNoIndex ){
          pElem = Tcl_NewStringObj((char*)cur2.name(), -1);
        } else {
          pElem = Tcl_NewStringObj((char*)cur2.basename(), -1);
        }
        Tcl_ListObjAppendElement(interp, pList, pElem);
      }
      Tcl_SetObjResult(interp, pList);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR memused ?PATH?
    **
    ** Return a list of 11 numbers that describe the number of nodes and
    ** the amount of memory allocated for the node PATH and all its
    ** decendents.
    */
    case XML_MEMUSED: {
      Tcl_Obj *pList;
      int i;
      int aMem[11];
      XmlCursor cur2 = *pCur;
      if( objc>=3 ){
        cur2.move(Tcl_GetStringFromObj(objv[2],0));
        if( !cur2.ok() ) break;
      }
      memset(aMem, 0, sizeof(aMem));
      cur2.memUsed(aMem);
      pList = Tcl_NewListObj(0, 0);
      for(i=0; i<(int)(sizeof(aMem)/sizeof(aMem[0])); i++){
        Tcl_Obj *pElem = Tcl_NewIntObj(aMem[i]);
        Tcl_ListObjAppendElement(interp, pList, pElem);
      }
      Tcl_SetObjResult(interp, pList);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR mkdir PATH
    **
    ** Create a new XML node whose name is PATH.  (A colon and integer
    ** will be appended to this name to make it unique among its siblings.)
    ** The full pathname of the new node is returned.
    **
    ** This command is similar to "create" except that this command does
    ** not change the XML node that the cursor is pointing to.  The
    ** "create" command, on the other hand, makes the cursor point to
    ** the newly created node.
    */
    case XML_MKDIR: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "PATH");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      const char *zDir = cur2.create(Tcl_GetStringFromObj(objv[2], 0));
      if( zDir ) Tcl_AppendElement(interp, zDir);
      break;
    }
    /*
    ** xmlp-tcl-cmd: CURSOR mv ?SRC? DEST
    **
    ** Move the subtree identified by SRC over to DEST.  If SRC is omitted,
    ** use the subtree that the cursor is currently pointing to.
    **
    ** If DEST does not previously exist, then SRC is renamed as DEST.
    ** If DEST does already exist, then SRC becomes a subelement of DEST.
    ** (This is how the Unix "mv" command works.)  For example, if the
    ** the command is
    **
    **             xml mv /abc:5/xyz:11 /pqr
    **
    ** If "/pqr:3" already exists, then the element "/abc:5/xyz:11/mno:0"
    ** will be called "/pqr:3/xyz:11/mno:0".  But if no "/pqr" existed
    ** prior to this command, then "/abc:5/xyz:11/mno:0" will become
    ** "/pqr:0/mno:0".
    */
    case XML_MV: {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?SRC? DEST");
        return TCL_ERROR;
      }
      XmlCursor from, to;
      from = *pCur;
      to = *pCur;
      const char *zDest;
      if( objc==3 ){
        zDest = Tcl_GetStringFromObj(objv[2],0);
      }else{
        if( !from.move(Tcl_GetStringFromObj(objv[2],0)) ) break;
        zDest = Tcl_GetStringFromObj(objv[3],0);
      }
      if( strcmp(from.fullname(), "/")==0 ){
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "Can't move from the root.", NULL);
        return TCL_ERROR;
      }
      to.move(zDest);
      if( to.ok() ){
        if( !to.create(from.basename()) ) break;  // fails due to trace proc
      }else{
        if( !to.create(zDest) ) break;  // fails due to trace proc
      }
      if( from.is_ancestor(&to) ){
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "Can't move ", from.fullname(),
            " to its descendent ", to.fullname(), ".", NULL);
        return TCL_ERROR;
      }
      from.DeepCopy(&from, &to);
      from.unlink();
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR name ?PATH?
    **
    ** Return the root name of the node at PATH (or the node to which the
    ** cursor is pointing if PATH is omitted).  The root name omits the index
    ** number.  Compare this command with "fullname" (or "pwd") and "basename".
    */
    case XML_NAME: {
      if( objc>3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH?");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      if( objc==3 ){
        cur2.move(Tcl_GetStringFromObj(objv[2], 0));
        if( !cur2.ok() ) break;
      }
      Tcl_AppendResult(interp, cur2.name(), 0);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR next ?PATTERN?
    **
    ** Move the cursor so that it points to the next sibling node whose
    ** name matches PATTERN.  If PATTERN is omitted, move the cursor to
    ** the next node.  If there are no more sibling nodes (that is, if
    ** the cursor already points to the last sibling) then this command
    ** is a no-op.
    */
    case XML_NEXT: {
      XmlCursor cur2 = *pCur;
      int rc;
      pCur->move_next();
      if( objc>2 ){
        char *zPattern = Tcl_GetStringFromObj(objv[2],0);
        while( pCur->ok() && globCompare(pCur->name(), zPattern)==0 ){
          pCur->move_next();
        }
      }
      rc = pCur->ok();
      if( !rc ){ *pCur = cur2; }
      Tcl_SetIntObj(Tcl_GetObjResult(interp), rc);
      break;
    }

    /* xmlp-tcl-cmd: CURSOR ok
    **
    ** Return TRUE if the last cursor movement command was successful and
    ** FALSE if it was not.  This can be used to text for the termination
    ** of a loop over all siblings. For example:
    **
    **    for {cursor first xyz} {[cursor ok]} {cursor next xyz} {...}
    */
    case XML_OK: {
      Tcl_SetIntObj(Tcl_GetObjResult(interp), pCur->ok());
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR parse ?PATH? ?-only TAG? XML
    **
    ** Parse XML text and place it into the tree at PATH or at the location
    ** that the cursor is pointing to if PATH is omitted.  If the
    ** "-only TAG" option is included, only the top-level XML element
    ** named TAG will be parsed and any other XML text will
    ** be ignored.
    **
    ** The full path of the head of the new path will be returned.
    */
    case XML_PARSE: {
      if( objc<3 || objc>6 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH? ?-only TAG? XML");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      char *zXml;
      int nXml = 0;
      char *zOnly = 0;
      int i = 2;
      int createOk = 1;
      while( i<objc-1 ){
        int nPath;
        char *zPath = Tcl_GetStringFromObj(objv[i], &nPath);
        if( nPath==5 && strncmp(zPath, "-only", nPath)==0 ){
          zOnly = Tcl_GetStringFromObj(objv[i+1], 0);
          i += 2;
        }else{
          cur2.move(zPath);
          if( !cur2.ok() ){
            if( !cur2.create(zPath) ){  // Create failed due to trace proc
              createOk = 0;
              break;
            }
          }
          i++;
        }
      }
      if( createOk ){
        zXml = (char*)Tcl_GetByteArrayFromObj(objv[i], &nXml);
        cur2.parse(zXml, nXml, zOnly);
      }
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR real ?PATH? ?VALUE?
    **
    ** Return the floating point value of the XML node named PATH or
    ** the current XML node if PATH is omitted.  If PATH does not exist,
    ** then return VALUE.  If PATH is not a floating point number, return 0.0.
    */
    case XML_REALVALUE: {
      if( objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH? ?VALUE?");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      if( objc>=3 ){
        cur2.move(Tcl_GetStringFromObj(objv[2], 0));
        if( !cur2.ok() ){
          if( objc==4 ){
            Tcl_SetObjResult(interp,Tcl_DuplicateObj(objv[3]));
          }
          break;
        }
      }
      Tcl_AppendResult(interp, cur2.coerce_real(), 0);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR render ?-strip-toplevel? ?PATH?
    **
    ** Generate XML text that describes the tree rooted at PATH or the tree
    ** that the cursor is pointing to if PATH is omitted.
    */
    case XML_RENDER: {
      if( objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH? ?-strip-toplevel?");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      int bStrip = 0;
      for(int i=2; i<objc; i++){
        if( strcmp("-strip-toplevel", Tcl_GetStringFromObj(objv[i], NULL))==0 ){
          bStrip = 1;
        } else {
          cur2.move(Tcl_GetStringFromObj(objv[i], 0));
          if( !cur2.ok() ){
            break;
          }
        }
      }
      char *zResult = cur2.render(0, bStrip);
      Tcl_AppendResult(interp, zResult, 0);
      if( zResult ) ckfree(zResult);
      break;
    }

    /*
    ** xmlp-tcl-cmd:  CURSOR renumber ?PATH? ?FIRST? ?INCREMENT?
    **
    ** Renumber the indices of all children of the node PATH.  If PATH 
    ** is omitted, then it is assumed to be ".".  The first index used
    ** is FIRST or 1 if FIRST is omitted.  Indices increment by INCREMENT
    ** or by 1 if INCREMENT is omitted. 
    **
    ** The return value is the largest index of any node after the
    ** renumber is complete.
    */
    case XML_RENUMBER: {
      int iFirst = 1, nStep = 1, mxIdx;
      if( objc>5 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH? ?FIRST? ?INCREMENT?");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      if( objc>2 ){
        const char *zArg = Tcl_GetStringFromObj(objv[2],0);
        if( !isdigit(zArg[0]) ){
          cur2.move(zArg);
          if( cur2.ok() ){
            objc--;
            objv++;
          }
        }
      }
      if( objc>2 ){
        if( Tcl_GetIntFromObj(interp, objv[2], &iFirst) ) return TCL_ERROR;
        objc--;
        objv++;
      }
      if( objc>2 ){
        if( Tcl_GetIntFromObj(interp, objv[2], &nStep) ) return TCL_ERROR;
      }
      mxIdx = cur2.renumber(iFirst,nStep);
      Tcl_SetObjResult(interp, Tcl_NewIntObj(mxIdx));
      break;
    }

    /*
    ** xmlp-tcl-cmd:  CURSOR reset_counter
    **
    ** Reset the index auto-generator to 0.
    */
    case XML_RESET_COUNTER: {
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, NULL);
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      cur2.resetCounter();
      Tcl_SetResult(interp, (char *)"", TCL_STATIC);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR set ?PATH? VALUE
    **
    ** Change the value of the XML node at PATH to VALUE.  If PATH is
    ** omitted, then the value of the node that the cursor is pointing to
    ** is changed.  This command makes PATH a leaf node.  If PATH used to
    ** contain subcontent, that subcontent is deleted.
    */
    case XML_SET: {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH? VALUE");
        return TCL_ERROR;
      }
      XmlCursor cur2 = *pCur;
      char *zValue;
      int nValue;

      /* There is, apparently, a bug in Tcl that is causing the result
      ** of Tcl_GetByteArrayFromObj() to become invalid sometimes 
      ** if a Tcl proc is executed, which happens if traces fire.
      ** Work around this problem by copying the byte array before doing
      ** the create() and set().
      */
      char *zNew = 0;
      if( objc==4 ){
        zValue = (char*)Tcl_GetByteArrayFromObj(objv[3],&nValue);
        zNew = new char[nValue];
        memcpy(zNew, zValue, nValue);
        char *zPath = Tcl_GetStringFromObj(objv[2], 0);
        cur2.move(zPath);
        if( !cur2.ok() ){
          if( !cur2.create(zPath) ) break;  // fails due to trace proc
        }
      }else{
        zValue = (char*)Tcl_GetByteArrayFromObj(objv[2],&nValue);
        zNew = new char[nValue];
        memcpy(zNew, zValue, nValue);
      }
      cur2.set(zNew, nValue);
      delete [] zNew; zNew = 0;
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR sync ?SRC? DEST ?-no-toplevel-attr?
    **
    ** Make the XML tree at DEST be an exact copy of the tree at SRC. 
    ** If SRC is omitted it defaults to ".".  
    **
    ** This command is similar to
    **
    **     CURSOR unset DEST
    **     CURSOR copy SRC DEST
    **
    ** The difference is that any traces or bindings on unchanged elements
    ** of DEST are preserved by this command but would be deleted by the
    ** unset and copy.  Also, this command is faster.
    **
    ** If -no-toplevel-attr is given, SRC must also be given.  In this
    ** case, don't sync the top-level xml attributes in SRC to DEST.  Leave
    ** DEST's attributes (if any) alone.
    */
    case XML_SYNC: {
      if( objc<3 || objc>5 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?SRC? DEST ?-no-toplevel-attr?");
        return TCL_ERROR;
      }
      XmlCursor from, to;
      from = *pCur;
      to = *pCur;
      int bNoToplevelAttr = 0;
      if( objc==3 ){

        /* Can fail due to trace proc
        */
        if( !to.moveOrCreate(Tcl_GetStringFromObj(objv[2],0)) ) break;
      }else if( objc>=4 ){
        char *zDest = Tcl_GetStringFromObj(objv[3], 0);
        if( !from.move(Tcl_GetStringFromObj(objv[2],0)) ){
          to.unlink(zDest);
          break;
        }
        if( !to.moveOrCreate(zDest) ) break;  // fails due to trace proc
        if( objc==5 ){
          char* zFifth = Tcl_GetStringFromObj(objv[4], 0);
          if( strcmp(zFifth, "-no-toplevel-attr")!=0 ){
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "invalid arg \"", zFifth,
                "\": must be \"-no-toplevel-attr\"");
            return TCL_ERROR;
          }
          bNoToplevelAttr = 1;
        }
      }
      if( from!=to ){
        from.DeepSync(&from, &to, bNoToplevelAttr);
      }
      break;
    }

    /*
    ** xmlp-tcl-cmd:  CURSOR trace ?-self|-children|-all|-info? ?PATH? PROC
    **
    ** Make arrangements to invoke a callback Tcl procedure whenever an
    ** element of the XML tree changes.  The trace is set at PATH or at
    ** "." if PATH is omitted.  PROC is the Tcl command to invoke whenever
    ** the element a PATH changes.
    **
    ** If the -self option is present, only the specific node named by PATH
    ** is traced. If -children is present, only the children of PATH but
    ** not PATH itself or the* grandchildren of later decendents of PATH
    ** are traced.  If -all is present, then PATH and all its decendents
    ** are traced.
    */
    case XML_TRACE: {
      if( objc<3 || objc>5 ){
        Tcl_WrongNumArgs(interp, 2, objv, 
            "?-self|-children|-all|-info? ?PATH? PROC");
        return TCL_ERROR;
      }
      XmlCursor cur = *pCur;
      if( strcmp(Tcl_GetStringFromObj(objv[2], 0),"-info")==0 ){
        if( objc==4 ){
          cur.move(Tcl_GetStringFromObj(objv[3], 0));
          if( !cur.ok() ) break;
        }
        XmlTraceCallback xFunc = 0;
        TraceCallbackData *pUserData = 0;
        int flags = 0;
        for(int i=0; cur.getTrace(i,&xFunc,(void**)&pUserData,&flags); i++){
          char zBuf[100];
          int iBuf = 0;
          if( xFunc!=traceCallback ) continue;
          if( flags & XML_TRACE_SELF ){
            strcpy(zBuf, "-self");
            iBuf = strlen(zBuf);
          }
          if( flags & XML_TRACE_CHILDREN ){
            if( iBuf>0 ) zBuf[iBuf++] = ' ';
            strcpy(&zBuf[iBuf], "-children");
            iBuf += strlen(&zBuf[iBuf]);
          }
          if( flags & XML_TRACE_ALL ){
            if( iBuf>0 ) zBuf[iBuf++] = ' ';
            strcpy(&zBuf[iBuf], "-all");
            iBuf += strlen(&zBuf[iBuf]);
          }
          Tcl_AppendElement(interp, zBuf);
          Tcl_AppendElement(interp, pUserData->zCallback);
        }
        break;
      }
      int flags = 0;
      char *zProc;
      if( parseTraceCommandLine(interp, objc-2, objv+2, &cur, &flags, &zProc) ){
        return TCL_ERROR;
      }
      TraceCallbackData *pData = new TraceCallbackData(interp, zProc);
      cur.trace(traceCallback, pData, flags, traceDeleteProc);
      break;
    }

    /*
    ** xmlp-tcl-cmd:  CURSOR unlink ?PATH? VARNAME
    **
    ** Undo the linkage between the variable VARNAME and the XML node
    ** at PATH.  If PATH is omitted, then it defaults to ".".  See the
    ** "link" command for additional information.
    */
    case XML_UNLINK: {
      if( objc<3 || objc>4 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH? VARNAME");
        return TCL_ERROR;
      }
      XmlCursor cur = *pCur;
      char *zVarName;
      if( objc==4 ){
        char *zPath = Tcl_GetStringFromObj(objv[2], 0);
        cur.move(zPath);
        if( !cur.ok() ){
          Tcl_AppendResult(interp, "no such XML element: ", zPath, 0);
          return TCL_ERROR;
        }
        zVarName = Tcl_GetStringFromObj(objv[3], 0);
      }else{
        zVarName = Tcl_GetStringFromObj(objv[2], 0);
      }
      XmlToTclLink *pData = 0;
      XmlTraceCallback xFunc = 0;
      int flags = 0;
      int i = 0;
      while( cur.getTrace(i++, &xFunc, (void**)&pData, &flags) ){
        if( xFunc==XmlToTclLink::xmlChange && flags==XML_TRACE_SELF && 
            strcmp(pData->varname(),zVarName)==0 ){
          delete pData; pData = 0;
          break;
        }
      }
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR unset ?PATH?
    **
    ** Delete the XML node at PATH and all its decendents.
    ** If PATH is omitted it defaults to ".".
    */
    /*
    ** xmlp-tcl-cmd: CURSOR rmdir ?PATH?
    **
    ** Delete the XML node at PATH and all its decendents.
    ** If PATH is omitted it defaults to ".".
    */
    case XML_UNSET:
    case XML_RMDIR: {
      if( objc>3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?PATH?");
        return TCL_ERROR;
      }
      const char *zPath = objc==3 ? Tcl_GetStringFromObj(objv[2], 0) : ".";
      pCur->unlink(zPath);
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR untrace ?-self|-children|-all? ?PATH? PROC
    **
    ** Remove a trace on a node of the XML tree.  See the "trace"
    ** command for additional information.
    */
    case XML_UNTRACE: {
      if( objc<3 || objc>5 ){
        Tcl_WrongNumArgs(interp, 2, objv, "?-self|-children|-all? ?PATH? PROC");
        return TCL_ERROR;
      }
      XmlCursor cur = *pCur;
      int flags = 0;
      char *zProc;
      if( parseTraceCommandLine(interp, objc-2, objv+2, &cur, &flags, &zProc) ){
        return TCL_ERROR;
      }
      TraceCallbackData *pData;
      XmlTraceCallback xFunc;
      int gotFlags;
      int i = 0;
      while( cur.getTrace(i++, &xFunc, (void**)&pData, &gotFlags) ){
        if( xFunc==traceCallback && gotFlags==flags && 
            strcmp(pData->zCallback,zProc)==0 ){
          cur.untrace(xFunc, pData, gotFlags);
          break;
        }
      }
      break;
    }
    default: {
      Tcl_AppendResult(interp, "this method has not yet been implemented", 0);
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}

/*
** This routine processes the "attr" method of an xml cursor. This
** routine is called by XmlCmdProc when the first argument is "attr".
*/
static int AttrCmdProc(
  Tcl_Interp *interp,     //Interpreter controlling the cursor
  int objc,               //Number of arguments
  Tcl_Obj *const* objv,   //Value of each argument
  XmlCursor *pCur         //Pointer to the cursor object
){
  static const char *ATTR_strs[] = {
    "boolean",            "exists",            "first",
    "get",                "ls",                "name",
    "next",               "ok",                "set",
    "unset",
    0
  };
  enum ATTR_enum {
    ATTR_BOOLEAN,         ATTR_EXISTS,          ATTR_FIRST,
    ATTR_GET,             ATTR_LS,              ATTR_NAME,
    ATTR_NEXT,            ATTR_OK,              ATTR_SET,
    ATTR_UNSET,
  };
  int index;
  int rc = TCL_OK;

  if( Tcl_GetIndexFromObj(interp, objv[2], ATTR_strs, "option", 0, &index)){
    return TCL_ERROR;
  }
  switch( (enum ATTR_enum)index ){
    /*
    ** xmlp-tcl-cmd: CURSOR attr boolean ?PATH? NAME ?VALUE?
    **
    ** Without a VALUE argument, this method returns true or false.  It returns
    ** true if PATH@NAME exists and is not zero.  It returns false if 
    ** PATH@NAME does not exist or has a value of zero.
    **
    ** If the argument VALUE exists and is true, then PATH@NAME is set to 1.
    ** If the argument VALUE exists and is false, then PATH@NAME is deleted.
    */
    case ATTR_BOOLEAN: {
      if( objc<4 || objc>6 ){
        Tcl_WrongNumArgs(interp, 3, objv, "?PATH? NAME ?VALUE?");
        return TCL_ERROR;
      }
      XmlCursor cur = *pCur;
      const char *zPath = NULL;
      const char *zName = NULL;
      int v = -1;
      switch( objc ){
        case 5: 

          /* Handle ambiguous command line.  If argument 4 is not a boolean,
          ** then this is: CURSOR attr boolean PATH NAME.
          ** If argument 4 is a boolean, then this is:
          ** CURSOR attr boolean NAME VALUE.
          */
          if( Tcl_GetBooleanFromObj(interp, objv[4], &v)!=TCL_OK ){
            zPath = Tcl_GetStringFromObj(objv[3], 0);
            zName = Tcl_GetStringFromObj(objv[4], 0);
            if( zPath ){ 
              cur.move(zPath);
              if( cur.ok() ){ 
                const char* z = cur.getAttribute(zName);
                int r = z!=0 && z[0]!=0 && z[0]!='0';
                Tcl_SetObjResult(interp, Tcl_NewBooleanObj(r));
              }
            }
            return TCL_OK; 
          }
          /* fall through, with v set to objv[4] */
        case 4: 
          zName = Tcl_GetStringFromObj(objv[3], 0);
          break;
        case 6:
          zPath = Tcl_GetStringFromObj(objv[3], 0);
          zName = Tcl_GetStringFromObj(objv[4], 0);
          if( Tcl_GetBooleanFromObj(interp, objv[5], &v) ){ return TCL_ERROR; }
          break;
      }
      if( objc==4 ){
        const char* z = cur.getAttribute(zName);
        int r = z!=0 && z[0]!=0 && z[0]!='0';
        Tcl_SetObjResult(interp, Tcl_NewBooleanObj(r));
      }else{
        if( v==0 ){
          if( zPath ){ cur.move(zPath); }
          if( cur.ok() ){ cur.removeAttribute(zName); }
        }else{
          if( zPath ){
            cur.move(zPath); 
            if( !cur.ok() ){
              if( !cur.create(zPath) ) break;  // fails due to trace proc
            }
          }
          cur.addAttribute(zName, "1");
        }
      }
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR attr exists ?PATH? NAME
    **
    ** Return true if the attribute named NAME exists for the XML element
    ** located at PATH. If PATH isn't specified, then the XML path that
    ** CURSOR is currently referencing is used.
    */
    case ATTR_EXISTS: {
      char *zPath = 0;
      char *zName = 0;
      XmlCursor cur2 = *pCur;

      if( objc==4 ){
        zName = Tcl_GetStringFromObj(objv[3], 0);
      }else if( objc==5 ){
        zPath = Tcl_GetStringFromObj(objv[3], 0);
        zName = Tcl_GetStringFromObj(objv[4], 0);
      }else{
        Tcl_WrongNumArgs(interp, 3, objv, "?PATH? NAME");
        return TCL_ERROR;
      }

      if( zPath ){
        cur2.move(zPath);
        if( !cur2.ok() ){
          Tcl_SetIntObj(Tcl_GetObjResult(interp),0);
          break;
        }
      }
      if( cur2.attr_exists(zName) ){
        Tcl_SetIntObj(Tcl_GetObjResult(interp),1);
      }else{ Tcl_SetIntObj(Tcl_GetObjResult(interp),0); }
      
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR attr first
    **
    ** Move the cursor so that it points to the first attribute.
    ** This method can be used in combination with the "next" method to
    ** implement a loop over all attributes of an XML node.
    */
    case ATTR_FIRST: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "");
        return TCL_ERROR;
      }
      pCur->first_attr();
      Tcl_SetIntObj(Tcl_GetObjResult(interp), pCur->attr_ok());
      break;
    }
    
    /*
    ** xmlp-tcl-cmd: CURSOR attr get ?PATH? ?NAME? ?VALUE?
    **
    ** Not your normal syntax:
    ** CURSOR attr get
    ** Return the value of the attribute currently being visited.
    **
    ** CURSOR attr get NAME
    ** Return the value of the attribute named NAME for the XML element that
    ** is currently referenced by CURSOR.
    **
    ** CURSOR attr get PATH NAME
    ** Return the value of the attribute named NAME for the XML element
    ** located at PATH.
    **
    ** CURSOR attr get PATH NAME VALUE
    ** Return the value of the attribute named NAME for the XML element
    ** located at PATH. If the attribute doesn't exist, return VALUE.
    */
    case ATTR_GET: {
      char *zPath = 0;
      char *zName = 0;
      char *zValue = 0;
      const char *temp = 0;
      XmlCursor cur2 = *pCur;

      if( objc==4 ){
        zName = Tcl_GetStringFromObj(objv[3], 0);
      }else if( objc==5 ){
        zPath = Tcl_GetStringFromObj(objv[3], 0);
        zName = Tcl_GetStringFromObj(objv[4], 0);
      }else if( objc==6 ){
        zPath = Tcl_GetStringFromObj(objv[3], 0);
        zName = Tcl_GetStringFromObj(objv[4], 0);
        zValue = Tcl_GetStringFromObj(objv[5], 0);
      }else if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "?PATH? ?NAME? ?VALUE?");
        return TCL_ERROR;
      }

      if( zPath ){
        cur2.move(zPath);
        if( !cur2.ok() ){ break; }
      }
      temp = cur2.getAttribute(zName);
      if( !temp ){
        if( zValue ){
          temp = (const char*)zValue;
        }else{
          temp = "";
        }
      }
      Tcl_Obj *pObj = Tcl_NewByteArrayObj(
              (const unsigned char*)temp,
              strlen(temp));
      Tcl_SetObjResult(interp, pObj);
      break;
    }
    
    /*
    ** xmlp-tcl-cmd: CURSOR attr ls ?PATH?
    **
    ** Return a list of attributes defined for the XML element at PATH.
    ** If PATH isn't specified, then the XML path that CURSOR is currently
    ** referencing is used.
    */
    case ATTR_LS: {
      char *zPath = 0;
      XmlCursor cur2 = *pCur;

      if( objc==4 ){
        zPath = Tcl_GetStringFromObj(objv[3], 0);
      }else if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "?PATH?");
        return TCL_ERROR;
      }

      if( zPath ){
        cur2.move(zPath);
        if( !cur2.ok() ){ break; }
      }
      if( !cur2.first_attr() ){ break; }
      Tcl_Obj *pObj = Tcl_NewByteArrayObj(
              (const unsigned char*)cur2.getAttrName(),
              strlen(cur2.getAttrName()));
      Tcl_SetListObj(Tcl_GetObjResult(interp), 1, &pObj);
      while( cur2.next_attr() ){
        pObj = Tcl_NewByteArrayObj(
                (const unsigned char*)cur2.getAttrName(),
                strlen(cur2.getAttrName()));
        Tcl_ListObjAppendElement(interp, Tcl_GetObjResult(interp), pObj);
      }
      break;
    }
    
    /*
    ** xmlp-tcl-cmd: CURSOR attr name
    **
    ** Return the name of the attribute currently being visited.
    */
    case ATTR_NAME: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "");
        return TCL_ERROR;
      }
      Tcl_Obj *pObj = Tcl_NewByteArrayObj(
              (const unsigned char*)pCur->getAttrName(),
              strlen(pCur->getAttrName()));
      Tcl_SetObjResult(interp, pObj);
      break;
    }
    
    /*
    ** xmlp-tcl-cmd: CURSOR attr next
    **
    ** Visit the "next" attribute for the XML path that CURSOR is currently
    ** referencing. If there are no more attributes for the XML path, this
    ** command does nothing (but see "attr ok"). Attributes are not ordered
    ** in any way.
    */
    case ATTR_NEXT: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "");
        return TCL_ERROR;
      }
      pCur->next_attr();
      Tcl_SetIntObj(Tcl_GetObjResult(interp), pCur->attr_ok());
      break;
    }
    
    /*
    ** xmlp-tcl-command: CURSOR attr ok
    ** 
    ** Return 1 if CURSOR is currently visiting an attribute for the XML path
    ** that CURSOR is currently referencing, and 0 otherwise.
    */
    case ATTR_OK: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "");
        return TCL_ERROR;
      }
      Tcl_SetIntObj(Tcl_GetObjResult(interp), pCur->attr_ok());
      break;
    }
    
    /*
    ** xmlp-tcl-command: CURSOR attr set ?PATH? ?NAME? VALUE
    **
    ** Set the value VALUE for the attribute named NAME for the XML element
    ** at PATH. If PATH isn't specified, then the XML path that CURSOR is
    ** currently referencing is used. If NAME isn't specified, then the
    ** attribute that CURSOR is currently visiting is used (see "attr first",
    ** "attr next").
    */
    case ATTR_SET: {
      char *zPath = 0;
      char *zName = 0;
      char *zValue = 0;
      XmlCursor cur2 = *pCur;

      if( objc==4 ){
        zValue = Tcl_GetStringFromObj(objv[3], 0);
      }else if( objc==5 ){
        zName = Tcl_GetStringFromObj(objv[3], 0);
        zValue = Tcl_GetStringFromObj(objv[4], 0);
      }else if( objc==6 ){
        zPath = Tcl_GetStringFromObj(objv[3], 0);
        zName = Tcl_GetStringFromObj(objv[4], 0);
        zValue = Tcl_GetStringFromObj(objv[5], 0);
      }else{
        Tcl_WrongNumArgs(interp, 3, objv, "?PATH? ?NAME? VALUE");
        return TCL_ERROR;
      }

      if( zPath ){
        cur2.move(zPath);
        if( !cur2.ok() ){
          if( !cur2.create(zPath) ) break;  // fails due to trace proc
        }
      }
      if( cur2.addAttribute(zName, zValue) ){
        Tcl_SetIntObj(Tcl_GetObjResult(interp),1);
      }else{ 
        Tcl_SetIntObj(Tcl_GetObjResult(interp),0);
      }
      
      break;
    }
    
    /*
    ** xmlp-tcl-command: CURSOR attr unset ?PATH? NAME
    **
    ** Delete the attribute named NAME for the XML element located at PATH.
    ** If PATH isn't specified, then the XML path that CURSOR is currently
    ** referencing is used.
    */
    case ATTR_UNSET: {
      char *zPath = 0;
      char *zName = 0;
      XmlCursor cur2 = *pCur;

      if( objc==4 ){
        zName = Tcl_GetStringFromObj(objv[3], 0);
      }else if( objc==5 ){
        zPath = Tcl_GetStringFromObj(objv[3], 0);
        zName = Tcl_GetStringFromObj(objv[4], 0);
      }else{
        Tcl_WrongNumArgs(interp, 3, objv, "?PATH? NAME");
        return TCL_ERROR;
      }

      if( zPath ){
        cur2.move(zPath);
        if( !cur2.ok() ){
          Tcl_SetIntObj(Tcl_GetObjResult(interp),0);
          break;
        }
      }
      if( cur2.removeAttribute(zName) ){
        Tcl_SetIntObj(Tcl_GetObjResult(interp),1);
      }else{ Tcl_SetIntObj(Tcl_GetObjResult(interp),0); }
      
      break;
    }
  }
  return rc;
}

/*
** The TCL interpreter calls this routine when a Tcl command associated
** with an XML cursor is deleted.  This routine deletes the associated
** XML cursor.
*/
static void XmlDeleteProc(ClientData pData){
  XmlCursor *pCur = (XmlCursor*)pData;
  delete pCur; pCur = 0;
}

/*
** tclcmd: xmlcursor CURSOR-NAME ?PATH?
** title:  Create a new XML cursor
**
** This routine runs when the Tcl interpreter executes an "xmlcursor"
** command.  This creates a new XML cursor and an associated TCL command
** used to control that cursor.  The name of the new TCL command is
** given by the first argument.  A second optional argument is the
** initial pathname of the cursor.
*/
static int XmlCreateCursor(
  ClientData NotUsed,      // Not used 
  Tcl_Interp *interp,      // The interpreter controlling the new cursor
  int objc,                // Number of arguments
  Tcl_Obj *const* objv     // Value of each argument
){
  const char *zPath;
  char *zCursorName;
  XmlCursor *pCur;
  char zNameBuf[100];
  if( objc<1 || objc>3 ){
    Tcl_WrongNumArgs(interp, 1, objv, "?CURSOR-NAME? ?PATH?");
    return TCL_ERROR;
  }
  zPath = objc==3 ? Tcl_GetStringFromObj(objv[2], 0) : "/";
  pCur = new XmlCursor(zPath);
  if( pCur==0 ){
    Tcl_AppendResult(interp, "unable to create the new cursor", 0);
    return TCL_ERROR;
  }
  if( objc>=2 ){
    zCursorName = Tcl_GetStringFromObj(objv[1], 0);
  }
  if( objc<2 || zCursorName[0]==0 ){
    sprintf(zNameBuf,"-xml-%p-", pCur);
    zCursorName = zNameBuf;
  }
  Tcl_CreateObjCommand(interp, zCursorName, XmlCmdProc, (void*)pCur,
                       XmlDeleteProc);
  Tcl_AppendResult(interp, zCursorName, 0);
  return TCL_OK;
}

/*
** tclcmd: xml_test_dump PATH
** title:  Print internal data of the XML tree
**
** This routine is used for testing and debugging only.  Print a
** diagnostic dump of internal data structures for the XML tree
** rooted at PATH.
*/
static int XmlTestDump(
  ClientData NotUsed,      // Not used 
  Tcl_Interp *interp,      // The interpreter controlling the new cursor
  int objc,                // Number of arguments
  Tcl_Obj *const* objv     // Value of each argument
){
  char *zPath;
  XmlCursor cur;
  if( objc!=2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "PATH");
    return TCL_ERROR;
  }
  zPath = Tcl_GetStringFromObj(objv[1], 0);
  cur.move(zPath);
  cur.testDump();
  return TCL_OK;
}

/*
** tclcmd: xml_test_ref PATH
** title:  Return the number of cursors currently point to PATH
**
** This routine is used for testing and debugging only.  It returns
** the number of cursors pointing to the XML node at PATH.
*/
static int XmlTestRef(
  ClientData NotUsed,      // Not used 
  Tcl_Interp *interp,      // The interpreter controlling the new cursor
  int objc,                // Number of arguments
  Tcl_Obj *const* objv     // Value of each argument
){
  char *zPath;
  XmlCursor cur;
  if( objc!=2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "PATH");
    return TCL_ERROR;
  }
  zPath = Tcl_GetStringFromObj(objv[1], 0);
  cur.move(zPath);
  Tcl_SetIntObj(Tcl_GetObjResult(interp), cur.testRef()-1);
  return TCL_OK;
}

/*
** tclcmd: xml_test_trace_list PATH
** title:  Return a list of all traces on an XML node
**
** This routine is used for testing and debugging only.  It returns
** a list of all traces on the XML node named PATH. 
*/
static int XmlTestTraceList(
  ClientData NotUsed,      // Not used 
  Tcl_Interp *interp,      // The interpreter controlling the new cursor
  int objc,                // Number of arguments
  Tcl_Obj *const* objv     // Value of each argument
){
  char *zPath;
  XmlCursor cur;
  if( objc!=2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "PATH");
    return TCL_ERROR;
  }
  zPath = Tcl_GetStringFromObj(objv[1], 0);
  cur.move(zPath);
  XmlTraceCallback xFunc;
  TraceCallbackData *pUserData;
  int flags;
  for(int i=0; cur.getTrace(i,&xFunc,(void**)&pUserData,&flags); i++){
    if( xFunc==traceCallback ){
      Tcl_AppendElement(interp, pUserData->zCallback);
    }else if( xFunc==XmlToTclLink::xmlChange ){
      char zBuf[300];
      sprintf(zBuf, "# XmlToTclLink(%.200s)",
          ((XmlToTclLink*)pUserData)->varname());
      Tcl_AppendElement(interp, zBuf);
    }else{
      char zBuf[100];
      sprintf(zBuf, "# C++ procedure at 0x%p", pUserData);
      Tcl_AppendElement(interp, zBuf);
    }
    char zFlags[30];
    sprintf(zFlags, "%d", flags);
    Tcl_AppendElement(interp, zFlags);
  }
  return TCL_OK;
}

/*
** This is the only function in this file with external linkage.
** This routine registers some commands with the Tcl interpreter then returns.
*/
extern "C" int Xmlp_Init(Tcl_Interp *interp){
  Tcl_CreateObjCommand(interp, "xmlcursor", XmlCreateCursor, 0, 0);
  Tcl_CreateObjCommand(interp, "xml_test_dump", XmlTestDump, 0, 0);
  Tcl_CreateObjCommand(interp, "xml_test_ref", XmlTestRef, 0, 0);
  Tcl_CreateObjCommand(interp, "xml_test_trace_list", XmlTestTraceList, 0, 0);
  if( Xmlp_Hash_Init(interp)!=TCL_OK ){ return TCL_ERROR; }
  return TCL_OK;
}
